﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;

namespace OsakaUniv.RIWL.Koudo.Writimation
{
    public partial class frmCanvas : Form
    {
        private const long CLICKTHRESHOLD = 10000000;   //  0.1 second
        private float mouseX;
        private float mouseY;
        private bool mouseDown_;
        private bool mouseMove_;
        private bool anchorSelected_;
        private bool selectSelectedAnchor_;
        private DrawingData ddata_;
        private List<DrawingData> undoBuffer_;
        private long baseTime_;                     //  used for drawing(100 nano second)
        private long drawingTime_;                  //  used for drawing(100 nano second)
        private long stepTime_;                     //  used for getting point data(100 nano second)
        private long elapsedTime_;                  //  used for animation(100 nano second)
        private long mouseDownTime_;                //  used for MouseClick(100 nano second)
        private int canvasWidth_;
        private int canvasHeight_;
        private string fileName_;
        private bool suppressKeyPress_;
        private RectangleF tmpRect_;
        private decimal scaleAlter_;
        private decimal scaleCache_;
        private bool shiftKeyPressed_;
        private bool controlKeyPressed_;
        private bool altKeyPressed_;
        private Keys keyCode_;
        private float canvasOffsetX_;
        private float canvasOffsetY_;
        private MouseButtons mouseButton_;
        private long strokeTimeInterval_;           //  (100 nano second)

        #region constructor
        //
        //  constructor
        //

        public frmCanvas()
        {
            InitializeComponent();

            mouseX = 0F;
            mouseY = 0F;
            mouseDown_ = false;
            mouseMove_ = false;
            anchorSelected_ = false;
            selectSelectedAnchor_ = false;
            ddata_ = new DrawingData();
            undoBuffer_ = new List<DrawingData>();
            baseTime_ = 0L;
            drawingTime_ = 0L;
            stepTime_ = 0L;
            elapsedTime_ = -1L;
            mouseDownTime_ = 0L;
            canvasWidth_ = 640;
            canvasHeight_ = 480;
            fileName_ = "NoName";
            tmpRect_ = RectangleF.Empty;
            scaleAlter_ = 100;
            scaleCache_ = nudScale.Value;
            shiftKeyPressed_ = false;
            controlKeyPressed_ = false;
            altKeyPressed_ = false;
            canvasOffsetX_ = 0.0f;
            canvasOffsetY_ = 0.0f;
            mouseButton_ = MouseButtons.None;
            strokeTimeInterval_ = 500;
        }

        #endregion  //  constructor

        #region method
        //
        //  method
        //

        private void AdjustStrokeInterval()
        {
            frmStrokeTimeInterval frm = new frmStrokeTimeInterval();
            frm.StrokeTimeInterval = this.strokeTimeInterval_;
            if (frm.ShowDialog() == DialogResult.OK)
            {
                this.strokeTimeInterval_ = frm.StrokeTimeInterval;
                this.SetUndoData();
                this.ddata_.SetStrokeInterval(this.strokeTimeInterval_ * 10000);
                MessageBox.Show(
                    "The stroke time interval was set.",
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Information);
            }
            picCanvas.Focus();
        }

        private long CalcPointTime()
        {
            long pointTime = DateTime.Now.Ticks - this.drawingTime_ + this.baseTime_;
            //this.drawingTime_ = DateTime.Now.Ticks;
            return pointTime;
        }

        private bool CheckStepTime()
        {
            return ((DateTime.Now.Ticks - this.stepTime_) > (10000000 / nudFPS.Value));
        }

        private void ConvertToPictures()
        {
            frmConvert frm = new frmConvert();
            frm.ImageWidth = (int)nudClipWidth.Value;
            frm.ImageHeight = (int)nudClipHeight.Value;
            frm.FPS = (float)nudFPS.Value;
            frm.BackgroundImageDraw = (cmbBGAtReplay.SelectedIndex == 1);
            frm.TextDraw = (cmbBGAtReplay.SelectedIndex == 1);
            frm.ShowDialog();
            picCanvas.Focus();
        }

        private void Forward()
        {
            tmrCanvas.Enabled = false;
            elapsedTime_ = -1L;
            this.RedrawCanvas();
        }

        private void RedrawCanvas()
        {
            picCanvas.Focus();
            picCanvas.Invalidate();
        }

        private void ResetTime()
        {
            SetUndoData();
            this.ddata_.SetStartTime(1);
            MessageBox.Show(
                "Start time was set.",
                Application.ProductName,
                MessageBoxButtons.OK,
                MessageBoxIcon.Information);
            picCanvas.Focus();
        }

        public void ResizeObject()
        {
            frmResize frm = new frmResize();
            if (frm.ShowDialog() == DialogResult.OK)
            {
                float xRatio = frm.XRatio;
                float yRatio = frm.YRatio;
                this.SetUndoData();
                this.SetDrawingDataProperties();
                this.ddata_.FontSize *= (Math.Max(xRatio, yRatio) / 100);
                this.ddata_.ScaleObject((xRatio / 100), (yRatio / 100));
                this.ddata_.SetAnchorPoints();
                this.RedrawCanvas();
            }
            picCanvas.Focus();
        }

        private void Rewind()
        {
            elapsedTime_ = 0L;
            this.RedrawCanvas();
        }

        private void ScaleCanvas()
        {
            this.canvasOffsetX_ *= (float)(nudScale.Value / this.scaleCache_);
            this.canvasOffsetY_ *= (float)(nudScale.Value / this.scaleCache_);
            this.scaleCache_ = nudScale.Value;
            this.RedrawCanvas();
        }

        private void SetBaseTime()
        {
            this.baseTime_ = this.ddata_.GetMaxTime();
        }

        private void SetDrawingDataProperties()
        {
            this.ddata_.OffsetX = this.canvasOffsetX_ + picCanvas.ClientSize.Width / 2;
            this.ddata_.OffsetY = this.canvasOffsetY_ + picCanvas.ClientSize.Height / 2;
            //this.ddata_.Scale = (float)(100 / nudScale.Value);
            this.ddata_.PenSize = (int)nudPenSize.Value;
            this.ddata_.PenColor = picPenColor.BackColor;
            this.ddata_.Scale = (float)(nudScale.Value / 100);
        }

        private void SetUndoData()
        {
            DrawingData data = new DrawingData();
            this.ddata_.CopyTo(data);
            this.undoBuffer_.Add(data);
        }

        private void Undo()
        {
            if (this.undoBuffer_.Count > 0)
            {
                this.ddata_ = this.undoBuffer_[this.undoBuffer_.Count - 1];
                this.undoBuffer_.RemoveAt(this.undoBuffer_.Count - 1);
                if (chkEdit.Checked)
                {
                    this.ddata_.SetAnchorPoints();
                }
                this.SetBaseTime();
                this.RedrawCanvas();
            }
            picCanvas.Focus();
        }

        private void WriteImage()
        {
            try
            {
                SaveFileDialog dlg = new SaveFileDialog();
                dlg.AutoUpgradeEnabled = true;
                dlg.CheckPathExists = true;
                dlg.DefaultExt = "wmf";
                dlg.Filter = "*.wmf|*.wmf|*.emf|*.emf|*.*|*.*";
                if (this.fileName_.IndexOf(".") > 0)
                {
                    dlg.FileName = this.fileName_.Substring(0, this.fileName_.LastIndexOf("."));
                }
                else
                {
                    dlg.FileName = this.fileName_;
                }
                if (dlg.ShowDialog() == DialogResult.OK)
                {
                    string fileName = dlg.FileName;
                    this.ddata_.PenSize = (int)nudPenSize.Value;
                    this.ddata_.PenColor = picPenColor.BackColor;
                    this.ddata_.WriteImage(
                        fileName,
                        (int)nudClipWidth.Value,
                        (int)nudClipHeight.Value,
                        (cmbBGAtReplay.SelectedIndex == 1),
                        (cmbBGAtReplay.SelectedIndex == 1),
                        ImageFormat.Emf);
                }
                picCanvas.Focus();
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmCanvas.btnWriteData_Click]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
        }

        #endregion  //  method

        #region event
        //
        //  event
        //

        private void btnAdjustStrokeInterval_Click(object sender, EventArgs e)
        {
            this.AdjustStrokeInterval();
            picCanvas.Focus();
        }

        private void btnAlignCenteredVertical_Click(object sender, EventArgs e)
        {
            this.ddata_.AlignSelectedObjectsCenteredVertical();
            this.RedrawCanvas();
        }

        private void btnAlignCenteredHorizontal_Click(object sender, EventArgs e)
        {
            this.ddata_.AlignSelectedObjectsCenteredHorizontal();
            this.RedrawCanvas();
        }

        private void btnBackgroundImage_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = "*.png|*.png|*.bmp|*.bmp|*.jpg|*.jpg|*.*|*.*";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                SetUndoData();
                Bitmap bmp = new Bitmap(dlg.FileName);
                this.ddata_.BackgroundImage = (Bitmap)bmp.Clone();
                this.RedrawCanvas();
            }
        }

        private void btnClear_Click(object sender, EventArgs e)
        {
            SetUndoData();
            this.ddata_.Points.Clear();
            this.baseTime_ = 0L;
            this.drawingTime_ = 0L;
            this.elapsedTime_ = -1L;
            this.RedrawCanvas();
        }

        private void btnClearBackground_Click(object sender, EventArgs e)
        {
            SetUndoData();
            this.ddata_.BackgroundImage = null;
            this.RedrawCanvas();
        }

        private void btnConvert_Click(object sender, EventArgs e)
        {
            this.ConvertToPictures();
        }

        private void btnDeletePoint_Click(object sender, EventArgs e)
        {
            if (!chkEdit.Checked)
            {
                return;
            }

            SetUndoData();
            this.ddata_.DeleteSelectedAnchorPoints();
            this.SetBaseTime();
            this.drawingTime_ = 0;
            this.RedrawCanvas();
        }

        private void btnForward_Click(object sender, EventArgs e)
        {
            this.Forward();
        }

        private void btnMove_Click(object sender, EventArgs e)
        {
            frmMove frm = new frmMove();

            RectangleF rect = this.ddata_.GetRectangle();
            frm.X = 0.0F;
            frm.Y = 0.0F;
            if (frm.ShowDialog() == DialogResult.OK)
            {
                SetUndoData();
                this.ddata_.MoveTo(frm.X, frm.Y);
                this.ddata_.SetAnchorPoints();
                this.RedrawCanvas();
            }
            picCanvas.Focus();
        }

        private void btnPause_Click(object sender, EventArgs e)
        {
            tmrCanvas.Enabled = false;
            this.RedrawCanvas();
        }

        private void btnPlay_Click(object sender, EventArgs e)
        {
            chkEdit.Checked = false;

            if (this.elapsedTime_ < 0)
            {
                elapsedTime_ = 0L;
            }

            tmrCanvas.Interval = (int)(1000 / (int)nudFPS.Value);
            tmrCanvas.Enabled = true;
        }

        private void btnReadData_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog dlg = new OpenFileDialog();
                dlg.Filter = "*.wrn|*.wrn";
                if (dlg.ShowDialog() == DialogResult.OK)
                {
                    this.fileName_ = dlg.FileName;
                    this.Text = this.fileName_;

                    SetUndoData();
                    this.ddata_.ReadData(dlg.FileName);
                    this.SetBaseTime();
                    this.drawingTime_ = 0L;
                    nudPenSize.Value = this.ddata_.PenSize;
                    chkEdit.Checked = false;
                    picPenColor.BackColor = this.ddata_.PenColor;
                    this.RedrawCanvas();
                }
                picCanvas.Focus();
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmCanvas.btnReadData_Click]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
        }

        private void btnResetTime_Click(object sender, EventArgs e)
        {
            this.ResetTime();
        }

        private void btnResize_Click(object sender, EventArgs e)
        {
            this.ResizeObject();
        }

        private void btnRewind_Click(object sender, EventArgs e)
        {
            this.Rewind();
        }

        private void btnSaveSequence_Click(object sender, EventArgs e)
        {
            try
            {
                SaveFileDialog dlg = new SaveFileDialog();
                dlg.AddExtension = false;
                dlg.Filter = "*.png|*.png|*.bmp|*.bmp|*.jpg|*.jpg|*.*|*.*";
                if (this.fileName_.LastIndexOf(".") > 0)
                {
                    dlg.FileName = this.fileName_.Substring(0, this.fileName_.LastIndexOf("."));
                }
                else
                {
                    dlg.FileName = this.fileName_;
                }
                if (dlg.ShowDialog() != DialogResult.OK)
                {
                    return;
                }

                this.Cursor = Cursors.WaitCursor;

                this.SetDrawingDataProperties();
                this.ddata_.WriteImages(
                    dlg.FileName,
                    (float)nudFPS.Value,
                    (int)nudClipWidth.Value,
                    (int)nudClipHeight.Value,
                    (cmbBGAtReplay.SelectedIndex == 1),
                    (cmbBGAtReplay.SelectedIndex == 1),
                    ImageFormat.Png);
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmCanvas.btnSaveSequence_Click]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
            finally
            {
                this.Cursor = Cursors.Default;
            }
        }

        private void btnText_Click(object sender, EventArgs e)
        {
            frmText frm = new frmText();
            frm.TextData = this.ddata_.Text;
            frm.FontDisplay = this.ddata_.Font;
            frm.TextColor = this.ddata_.TextColor;
            if (frm.ShowDialog() == DialogResult.OK)
            {
                this.ddata_.Text = frm.TextData;
                this.ddata_.Font = frm.FontDisplay;
                this.ddata_.TextColor = frm.TextColor;
                this.RedrawCanvas();
            }
            picCanvas.Focus();
        }

        private void btnUndo_Click(object sender, EventArgs e)
        {
            this.Undo();
        }

        private void btnWriteData_Click(object sender, EventArgs e)
        {
            try
            {
                SaveFileDialog dlg = new SaveFileDialog();
                dlg.AutoUpgradeEnabled = true;
                dlg.CheckPathExists = true;
                dlg.DefaultExt = "wrn";
                dlg.Filter = "*.wrn|*.wrn|*.*|*.*";
                dlg.FileName = this.fileName_;
                if (dlg.ShowDialog() == DialogResult.OK)
                {
                    this.fileName_ = dlg.FileName;

                    this.ddata_.PenSize = (int)nudPenSize.Value;
                    this.ddata_.PenColor = picPenColor.BackColor;
                    this.ddata_.WriteData(this.fileName_);
                    this.Text = this.fileName_;
                }
                picCanvas.Focus();
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmCanvas.btnWriteData_Click]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
        }

        private void btnWriteImage_Click(object sender, EventArgs e)
        {
            this.WriteImage();
        }

        private void chkCircle_CheckedChanged(object sender, EventArgs e)
        {
            this.tmpRect_ = RectangleF.Empty;
            this.RedrawCanvas();
        }

        private void chkEdit_CheckedChanged(object sender, EventArgs e)
        {
            if (chkEdit.Checked)
            {
                this.ddata_.SetAnchorPoints();
            }
            else
            {
                chkCircle.Checked = false;
                this.ddata_.AnchorPoints.Clear();
                this.SetBaseTime();
                this.drawingTime_ = 0L;
            }
            this.RedrawCanvas();
        }

        private void frmCanvas_Click(object sender, EventArgs e)
        {

        }

        private void frmCanvas_FormClosed(object sender, FormClosedEventArgs e)
        {
        }

        private void frmCanvas_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this.ddata_.GetPointCount() > 0)
            {
                if (MessageBox.Show(
                    "Quit without saving.\r\nAre you sure?",
                    Application.ProductName,
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Question,
                    MessageBoxDefaultButton.Button2) == DialogResult.No)
                {
                    e.Cancel = true;
                    return;
                }
            }
        }

        private void frmCanvas_Load(object sender, EventArgs e)
        {
            this.Text = this.fileName_;
            lblElapsedTime.Text = "";
            if (cmbBGAtReplay.Items.Count > 0)
            {
                cmbBGAtReplay.SelectedIndex = 1;
            }
            nudClipWidth.Value = (decimal)this.canvasWidth_;
            nudClipHeight.Value = (decimal)this.canvasHeight_;
        }

        private void frmCanvas_KeyDown(object sender, KeyEventArgs e)
        {
            this.suppressKeyPress_ = true;
            e.SuppressKeyPress = true;
            this.shiftKeyPressed_ = e.Shift;
            this.controlKeyPressed_ = e.Control;
            this.altKeyPressed_ = e.Alt;
            this.keyCode_ = e.KeyCode;

            this.SetDrawingDataProperties();

            switch (e.KeyCode)
            {
                case Keys.Delete:
                    if (chkEdit.Checked)
                    {
                        this.btnDeletePoint_Click(null, null);
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                //case Keys.Space:
                //    if (tmrCanvas.Enabled)
                //    {
                //        this.btnPause_Click(null, null);
                //    }
                //    else
                //    {
                //        this.btnPlay_Click(null, null);
                //    }
                //    break;
                case Keys.Enter:
                    picCanvas.Focus();
                    break;
                case Keys.Up:
                    if (picCanvas.Focused
                        && chkEdit.Checked
                        && this.ddata_.IsSelectedAnchorExist())
                    {
                        this.ddata_.MoveSelectedPointTo(0, (float)(-100 / nudScale.Value));
                        this.RedrawCanvas();
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.Down:
                    if (picCanvas.Focused
                        && chkEdit.Checked
                        && this.ddata_.IsSelectedAnchorExist())
                    {
                        this.ddata_.MoveSelectedPointTo(0, (float)(100 / nudScale.Value));
                        this.RedrawCanvas();
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.Left:
                    if (picCanvas.Focused
                        && chkEdit.Checked
                        && this.ddata_.IsSelectedAnchorExist())
                    {
                        this.ddata_.MoveSelectedPointTo((float)(-100 / nudScale.Value), 0);
                        this.RedrawCanvas();
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.Right:
                    if (picCanvas.Focused
                        && chkEdit.Checked
                        && this.ddata_.IsSelectedAnchorExist())
                    {
                        this.ddata_.MoveSelectedPointTo((float)(100 / nudScale.Value), 0);
                        this.RedrawCanvas();
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.Home:
                    this.Rewind();
                    break;
                case Keys.End:
                    this.Forward();
                    break;
                case Keys.D0:
                    if (e.Control)
                    {
                        decimal tmpScale = nudScale.Value;
                        nudScale.Value = this.scaleAlter_;
                        this.scaleAlter_ = tmpScale;
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.A:
                    if (e.Control)
                    {
                        if (chkEdit.Checked)
                        {
                            if (e.Shift)
                            {
                                this.ddata_.UnselectAnchorAll();
                            }
                            else
                            {
                                this.ddata_.SelectAnchorAll();
                            }
                            this.RedrawCanvas();
                        }
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.E:
                    if (e.Control)
                    {
                        chkEdit.Checked = !chkEdit.Checked;
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.M:
                    if (e.Control)
                    {
                        this.btnMove_Click(null, null);
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.O:
                    if (e.Control)
                    {
                        this.btnReadData_Click(null, null);
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.S:
                    if (e.Control)
                    {
                        this.btnWriteData_Click(null, null);
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.T:
                    if (e.Control)
                    {
                        this.btnText_Click(null, null);
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.W:
                    if (e.Control)
                    {
                        this.Close();
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.Z:
                    if (e.Control)
                    {
                        this.Undo();
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.Oemplus:
                case Keys.Add:
                    if (e.Control)
                    {
                        if (nudScale.Maximum >= nudScale.Value + nudScale.Increment)
                        {
                            nudScale.Value += nudScale.Increment;
                        }
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                case Keys.OemMinus:
                case Keys.Subtract:
                    if (e.Control)
                    {
                        if (nudScale.Minimum <= nudScale.Value - nudScale.Increment)
                        {
                            nudScale.Value -= nudScale.Increment;
                        }
                    }
                    else
                    {
                        goto default;
                    }
                    break;
                default:
                    this.suppressKeyPress_ = false;
                    e.SuppressKeyPress = false;
                    break;
            }
        }

        private void frmCanvas_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Space:
                    if (this.mouseMove_)
                    {
                        goto default;
                    }
                    else if (tmrCanvas.Enabled)
                    {
                        this.btnPause_Click(null, null);
                    }
                    else
                    {
                        this.btnPlay_Click(null, null);
                    }
                    break;
                default:
                    break;
            }

            this.shiftKeyPressed_ = e.Shift;
            this.controlKeyPressed_ = e.Control;
            this.altKeyPressed_ = e.Alt;
            this.keyCode_ = e.KeyCode;
            this.suppressKeyPress_ = false;
        }

        private void nudClipHeight_KeyDown(object sender, KeyEventArgs e)
        {
            e.SuppressKeyPress = this.suppressKeyPress_;
        }

        private void nudClipHeight_KeyPress(object sender, KeyPressEventArgs e)
        {
        }

        private void nudClipHeight_ValueChanged(object sender, EventArgs e)
        {
            this.canvasHeight_ = (int)nudClipHeight.Value;
            this.RedrawCanvas();
        }

        private void nudClipWidth_KeyDown(object sender, KeyEventArgs e)
        {
            e.SuppressKeyPress = this.suppressKeyPress_;
        }

        private void nudClipWidth_KeyPress(object sender, KeyPressEventArgs e)
        {
        }

        private void nudClipWidth_ValueChanged(object sender, EventArgs e)
        {
            this.canvasWidth_ = (int)nudClipWidth.Value;
            this.RedrawCanvas();
        }

        private void nudFPS_KeyDown(object sender, KeyEventArgs e)
        {
            e.SuppressKeyPress = this.suppressKeyPress_;
        }

        private void nudFPS_KeyPress(object sender, KeyPressEventArgs e)
        {
        }

        private void nudPenSize_KeyDown(object sender, KeyEventArgs e)
        {
            e.SuppressKeyPress = this.suppressKeyPress_;
        }

        private void nudPenSize_KeyPress(object sender, KeyPressEventArgs e)
        {
            e.Handled = this.suppressKeyPress_;
        }

        private void nudPenSize_ValueChanged(object sender, EventArgs e)
        {
            this.RedrawCanvas();
        }

        private void nudScale_KeyDown(object sender, KeyEventArgs e)
        {
            //e.Handled = this.suppressKeyPress_;
            e.SuppressKeyPress = this.suppressKeyPress_;
            switch (e.KeyCode)
            {
                case Keys.Left:
                case Keys.Right:
                case Keys.Up:
                case Keys.Down:
                    break;
            }
        }

        private void nudScale_KeyPress(object sender, KeyPressEventArgs e)
        {
            e.Handled = this.suppressKeyPress_;
            switch (e.KeyChar)
            {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                case (char)Keys.Back:
                    break;
                default:
                    e.Handled = true;
                    break;
            }
        }

        private void nudScale_KeyUp(object sender, KeyEventArgs e)
        {
            e.Handled = this.suppressKeyPress_;
        }

        private void nudScale_ValueChanged(object sender, EventArgs e)
        {
            this.ScaleCanvas();
        }

        private void picCanvas_Click(object sender, EventArgs e)
        {
            try
            {
                if (chkEdit.Checked)
                {
                    if ((DateTime.Now.Ticks - this.mouseDownTime_) < CLICKTHRESHOLD)
                    {
                        return;
                    }
                    //
                    //  Maybe value of mouseX and mouseY are setting in the "picCanvas_MouseUp()".
                    //
                    //this.ddata_.OffsetX = picCanvas.ClientSize.Width / 2;
                    //this.ddata_.OffsetY = picCanvas.ClientSize.Height / 2;
                    //this.ddata_.Scale = (float)(100 / nudScale.Value);
                    this.SetDrawingDataProperties();
                    if (this.ddata_.UnselectAnchor(mouseX, mouseY) >= 0)
                    {
                        this.RedrawCanvas();
                    }
                }
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmCanvas.picCanvas_Click]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
            finally
            {
                this.mouseDown_ = false;
                this.mouseMove_ = false;
            }
        }

        private void picCanvas_DoubleClick(object sender, EventArgs e)
        {
            switch (this.mouseButton_)
            {
                case MouseButtons.Right:
                    this.canvasOffsetX_ = 0.0f;
                    this.canvasOffsetY_ = 0.0f;
                    this.RedrawCanvas();
                    break;
            }
        }

        private void picCanvas_MouseDown(object sender, MouseEventArgs e)
        {
            this.mouseX = (float)e.X;
            this.mouseY = (float)e.Y;
            this.mouseButton_ = e.Button;

            switch (e.Button)
            {
                case MouseButtons.Left:
                    this.mouseDown_ = true;
                    this.mouseMove_ = false;
                    this.mouseDownTime_ = DateTime.Now.Ticks;
                    this.elapsedTime_ = -1L;
                    //this.SetBaseTime();
                    if ((this.drawingTime_ == 0)
                        || (this.ddata_.GetPointCount() <= 0))
                    {
                        this.drawingTime_ = DateTime.Now.Ticks;
                    }
                    else
                    {
                        long maxTime = this.ddata_.GetMaxTime();
                        if (DateTime.Now.Ticks - this.drawingTime_ - maxTime > (this.strokeTimeInterval_ * 10000))
                        {
                            this.baseTime_ = maxTime + (this.strokeTimeInterval_ * 10000);
                            this.drawingTime_ = DateTime.Now.Ticks;
                        }
                    }
                    try
                    {
                        SetUndoData();

                        //this.ddata_.OffsetX = picCanvas.ClientSize.Width / 2;
                        //this.ddata_.OffsetY = picCanvas.ClientSize.Height / 2;
                        //this.ddata_.Scale = (float)(100 / nudScale.Value);
                        this.SetDrawingDataProperties();

                        if (chkEdit.Checked)
                        {
                            this.selectSelectedAnchor_ = this.ddata_.IsAnchorSelected(mouseX, mouseY);
                            this.anchorSelected_ = false;
                            int n = this.ddata_.SelectAnchor(mouseX, mouseY);
                            if (n >= 0)
                            {
                                PointData pd = this.ddata_.AnchorPoints.Points[n];
                                lblElapsedTime.Text =
                                    "[" + n.ToString() + "]"
                                    + "(" + pd.X.ToString() + ", "
                                    + pd.Y.ToString() + ")";
                                this.anchorSelected_ = true;
                                this.RedrawCanvas();
                            }
                        }
                        else
                        {
                            this.ddata_.Add(
                                mouseX,
                                mouseY,
                                this.CalcPointTime());
                            this.stepTime_ = DateTime.Now.Ticks;
                            //Console.WriteLine("MouseDown: AddPoint");
                        }
                    }
                    catch (Exception exp)
                    {
                        MessageBox.Show(
                            "[frmCanvas.picCanvas_MouseDown]"
                            + "(" + exp.Source + ")\r\n"
                            + "  " + exp.Message,
                            Application.ProductName,
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Error);
                    }
                    break;
            }
        }

        private void picCanvas_MouseMove(object sender, MouseEventArgs e)
        {
            this.SetDrawingDataProperties();

            switch (e.Button)
            {
                case MouseButtons.Left:
                    try
                    {
                        this.mouseMove_ = true;
                        float x = (float)e.X;
                        float y = (float)e.Y;

                        if (chkEdit.Checked)
                        {
                            this.tmpRect_ = RectangleF.Empty;

                            if (this.anchorSelected_)
                            {
                                this.ddata_.MoveSelectedPointTo(
                                    x - mouseX,
                                    y - mouseY);
                                mouseX = x;
                                mouseY = y;
                            }
                            else
                            {
                                float left = mouseX;
                                float top = mouseY;
                                float right = x;
                                float bottom = y;
                                if (left > right)
                                {
                                    left = x;
                                    right = mouseX;
                                }
                                if (top > bottom)
                                {
                                    top = y;
                                    bottom = mouseY;
                                }
                                float width = Math.Abs(right - left);
                                float height = Math.Abs(bottom - top);
                                if (this.shiftKeyPressed_)
                                {
                                    if (width > height)
                                    {
                                        height = width;
                                    }
                                    else if (width < height)
                                    {
                                        width = height;
                                    }
                                }

                                this.tmpRect_ = new RectangleF(left, top, width, height);

                                if (chkCircle.Checked)
                                {
                                }
                                else
                                {
                                    if (this.ddata_.SelectAnchorInRegion(left, top, right, bottom) > 0)
                                    {
                                    }
                                }
                            }
                        }
                        else if (this.CheckStepTime())
                        {
                            this.ddata_.AddPoint(x, y, this.CalcPointTime());
                            this.stepTime_ = DateTime.Now.Ticks;
                        }
                        this.RedrawCanvas();
                    }
                    catch (Exception exp)
                    {
                        MessageBox.Show(
                            "[frmCanvas.picCanvas_MouseMove]"
                            + "(" + exp.Source + ")\r\n"
                            + "  " + exp.Message,
                            Application.ProductName,
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Error);
                    }
                    break;
                case MouseButtons.Right:
                    this.canvasOffsetX_ += e.X - this.mouseX;
                    this.canvasOffsetY_ += e.Y - this.mouseY;
                    //Console.WriteLine("({0}, {1}) ({2}, {3})", e.X, e.Y, this.mouseX, this.mouseY);
                    this.mouseX = (float)e.X;
                    this.mouseY = (float)e.Y;
                    this.RedrawCanvas();
                    break;
            }
        }

        private void picCanvas_MouseUp(object sender, MouseEventArgs e)
        {
            switch (e.Button)
            {
                case MouseButtons.Left:
                    mouseX = (float)e.X;
                    mouseY = (float)e.Y;
                    this.elapsedTime_ = -1L;
                    //if (this.drawingTime_ == 0)
                    //{
                        //this.drawingTime_ = DateTime.Now.Ticks;
                    //}
                    try
                    {
                        this.SetDrawingDataProperties();

                        if (chkEdit.Checked)
                        {
                            if (chkCircle.Checked)
                            {
                                if (this.tmpRect_ != RectangleF.Empty)
                                {
                                    this.SetUndoData();
                                    this.ddata_.TransformToEllipse(this.tmpRect_);
                                    this.ddata_.SetAnchorPoints();
                                }
                                chkCircle.Checked = false;
                            }
                            else if (this.selectSelectedAnchor_ && !mouseMove_)
                            {
                                if (this.ddata_.UnselectAnchor(mouseX, mouseY) >= 0)
                                {
                                }
                            }
                        }
                        else if (this.ddata_.GetPointsCount(this.ddata_.Points.Count - 1) == 1)
                        {
                            //Console.WriteLine("MouseUp: Before AddPoint({0})", this.ddata_.GetPointsCount(this.ddata_.Points.Count - 1));
                            this.ddata_.AddPoint(
                                (float)e.X,
                                (float)e.Y, 
                                this.CalcPointTime());
                            if (this.ddata_.GetPointsCount(this.ddata_.Points.Count - 1) == 1)
                            {
                                this.ddata_.AddPoint(
                                    (float)e.X + 1,
                                    (float)e.Y + 1,
                                    this.CalcPointTime());
                            }
                            this.stepTime_ = DateTime.Now.Ticks;
                            //Console.WriteLine("MouseUp: After AddPoint({0})", this.ddata_.GetPointsCount(this.ddata_.Points.Count - 1));
                        }
                        this.tmpRect_ = RectangleF.Empty;
                        this.RedrawCanvas();
                    }
                    catch (Exception exp)
                    {
                        MessageBox.Show(
                            "[frmCanvas.picCanvas_MouseUp]"
                            + "(" + exp.Source + ")\r\n"
                            + "  " + exp.Message,
                            Application.ProductName,
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Error);
                    }
                    finally
                    {
                        this.mouseDown_ = false;
                        //this.mouseMove_ = false;
                    }
                    break;
            }
        }

        private void picCanvas_Paint(object sender, PaintEventArgs e)
        {
            try
            {
                this.SetDrawingDataProperties();

                if ((this.elapsedTime_ < 0) 
                    || ((this.elapsedTime_ >= 0) && (cmbBGAtReplay.SelectedIndex == 1)))
                {
                    //  background image
                    this.ddata_.DrawImage(e.Graphics);

                    //  background text
                    if (this.ddata_.Text.Length > 0)
                    {
                        this.ddata_.DrawString(e.Graphics);
                    }
                }

                if (this.elapsedTime_ < 0)
                {
                    //  drawing draw area.
                    e.Graphics.DrawRectangle(
                        new Pen(Color.CadetBlue, 1),
                        (picCanvas.ClientSize.Width - this.canvasWidth_ * this.ddata_.Scale) / 2,
                        (picCanvas.ClientSize.Height - this.canvasHeight_ * this.ddata_.Scale) / 2,
                        this.canvasWidth_ * this.ddata_.Scale,
                        this.canvasHeight_ * this.ddata_.Scale);

                    //  drawing a ruled line.
                    e.Graphics.DrawLine(
                        new Pen(Color.CadetBlue, 1),
                        picCanvas.ClientSize.Width / 2,
                        0,
                        picCanvas.ClientSize.Width / 2,
                        picCanvas.ClientSize.Height);
                    e.Graphics.DrawLine(
                        new Pen(Color.CadetBlue, 1),
                        0,
                        picCanvas.ClientSize.Height / 2,
                        picCanvas.ClientSize.Width,
                        picCanvas.ClientSize.Height / 2);
                }

                //  drawing the input data.
                this.ddata_.DrawCurve(
                    e.Graphics,
                    picPenColor.BackColor,
                    (int)nudPenSize.Value,
                    this.elapsedTime_);

                //  drawing the anchor
                if (chkEdit.Checked)
                {
                    this.ddata_.DrawAnchorPoints(e.Graphics);
                }

                //  drawing the model
                if (tmpRect_ != RectangleF.Empty)
                {
                    if (chkCircle.Checked)
                    {
                        e.Graphics.DrawEllipse(
                            new Pen(Color.DarkGray),
                            this.tmpRect_);
                    }
                    else
                    {
                        e.Graphics.DrawRectangle(
                            new Pen(Color.Red),
                            this.tmpRect_.X,
                            this.tmpRect_.Y,
                            this.tmpRect_.Width,
                            this.tmpRect_.Height);
                    }
                }
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmCanvas.picCanvas_Paint]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
        }

        private void picPenColor_Click(object sender, EventArgs e)
        {
            ColorDialog dlg = new ColorDialog();
            dlg.Color = picPenColor.BackColor;
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                picPenColor.BackColor = dlg.Color;
                this.RedrawCanvas();
            }
        }

        private void tmrCanvas_Tick(object sender, EventArgs e)
        {
            tmrCanvas.Enabled = false;

            elapsedTime_ += tmrCanvas.Interval * 10000;
            lblElapsedTime.Text = elapsedTime_.ToString();
            lblElapsedTime.Refresh();

            if (elapsedTime_ > this.ddata_.GetMaxTime())
            {
                elapsedTime_ = -1;
            }
            else
            {
                tmrCanvas.Enabled = true;
            }

            this.RedrawCanvas();
        }

        #endregion  //  event

        private void button1_Click(object sender, EventArgs e)
        {
            Graphics g = Graphics.FromHwnd(picCanvas.Handle);
            g.DrawRectangle(new Pen(Color.Blue), 100 - 80, 100 - 60, 160, 120);
            for (int i = 0; i <= 360; i++)
            {
                double a = (double)i * Math.PI / 180;
                double x = 80 * Math.Cos(a);
                double y = 60 * Math.Sin(a);
                g.FillEllipse(new SolidBrush(Color.LightBlue), (float)(100 + x), (float)(100 + y), 5, 5);
                double aa = Math.Atan2(y, x);
                if ((x < 0) && (y < 0))
                {
                    aa = Math.Atan2(-y, -x) + Math.PI;
                }
                else if ((x >= 0) && (y < 0))
                {
                    aa = 2 * Math.PI + Math.Atan2(y, x);
                }
                double xx = 40 * Math.Cos(aa);
                double yy = 30 * Math.Sin(aa);
                g.FillEllipse(new SolidBrush(Color.Indigo), (float)(100 + xx), (float)(100 + yy), 2, 2);
                //Console.WriteLine("{0}, {1}, {2} -> {3}", i, x, y, aa * 180 / Math.PI);
            }
        }
    }
}
