﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace OsakaUniv.RIWL.Koudo.Writimation
{
    public partial class frmConvert : Form
    {
        private List<string> fileNames_;
        private float fps_;
        private int imageWidth_;
        private int imageHeight_;
        private float imageScale_;
        private bool textDraw_;
        private bool backgroundImageDraw_;
        private bool processing_;
        private bool cancel_;

        public frmConvert()
        {
            InitializeComponent();

            this.fileNames_ = new List<string>();
            this.fps_ = 30;
            this.imageWidth_ = 640;
            this.imageHeight_ = 480;
            this.imageScale_ = 1.0F;
            this.textDraw_ = false;
            this.backgroundImageDraw_ = false;
            this.processing_ = false;
            this.cancel_ = false;
        }

        #region property
        //
        //  property
        //

        public bool BackgroundImageDraw
        {
            get { return this.backgroundImageDraw_; }
            set { this.backgroundImageDraw_ = value; }
        }

        public float FPS
        {
            get { return this.fps_; }
            set { this.fps_ = value; }
        }

        public int ImageHeight
        {
            get { return this.imageHeight_; }
            set { this.imageHeight_ = value; }
        }

        public int ImageWidth
        {
            get { return this.imageWidth_; }
            set { this.imageWidth_ = value; }
        }

        public float ImageScale
        {
            get { return this.imageScale_; }
            set { this.imageScale_ = value; }
        }

        public bool TextDraw
        {
            get { return this.textDraw_; }
            set { this.textDraw_ = value; }
        }

        #endregion  //  property

        #region method
        //
        //  method
        //

        private void AddStringToListBox(
            ListBox lst,
            string s)
        {
            if (lst.Items.Count > 2000)
            {
                lst.Items.RemoveAt(0);
            }
            lst.Items.Add(
                "[" + DateTime.Now.ToString("yyyy.mm.dd hh:mm:ss") + "]" 
                + s);
            lst.SetSelected(lst.Items.Count - 1, true);
            lst.Refresh();
        }

        #endregion  //  method

        #region event
        //
        //  event
        //

        private void btnClose_Click(object sender, EventArgs e)
        {
            if (this.processing_)
            {
                this.cancel_ = true;
                //while (this.processing_)
                //{
                //    Application.DoEvents();
                //}
            }
            else
            {
                this.Close();
            }
        }

        private void btnConvert_Click(object sender, EventArgs e)
        {
            try
            {
                this.Cursor = Cursors.WaitCursor;

                this.cancel_ = false;
                this.processing_ = true;

                foreach (Control ictr in this.Controls)
                {
                    if (ictr is Button)
                    {
                        if (!ictr.Name.Equals("btnClose"))
                        {
                            ictr.Enabled = false;
                        }
                    }
                }

                DrawingData ddata_ = new DrawingData();
                for (int i = 0; i < this.fileNames_.Count; i++)
                {
                    Application.DoEvents();
                    if (this.cancel_)
                    {
                        this.AddStringToListBox(
                            lstMessage,
                            "> canceled.");
                        break;
                    }

                    ddata_.Clear();
                    ddata_.Scale = this.imageScale_;
                    ddata_.ReadData(this.fileNames_[i]);
                    if (ddata_.GetPointCount() <= 0)
                    {
                        continue;
                    }

                    lstSource.SetSelected(i, true);
                    this.AddStringToListBox(
                        lstMessage,
                        "start converting.");
                    this.AddStringToListBox(
                        lstMessage,
                        "file name: " + this.fileNames_[i]);
                    this.AddStringToListBox(
                        lstMessage,
                        "> processing...");

                    System.Drawing.Imaging.ImageFormat imgfmt = System.Drawing.Imaging.ImageFormat.Png;
                    if (rbtnPNG.Checked)
                    {
                        imgfmt = System.Drawing.Imaging.ImageFormat.Png;
                    }
                    else if (rbtnJPEG.Checked)
                    {
                        imgfmt = System.Drawing.Imaging.ImageFormat.Jpeg;
                    }

                    if (txtSavePath.Text == "")
                    {
                        ddata_.WriteImages(
                            this.fileNames_[i],
                            this.fps_,
                            this.imageWidth_,
                            this.imageHeight_,
                            this.textDraw_,
                            this.backgroundImageDraw_,
                            imgfmt);
                    }
                    else
                    {
                        string fileName = this.fileNames_[i];
                        if (fileName.LastIndexOf(@"\") >= 0)
                        {
                            fileName = fileName.Substring(fileName.LastIndexOf(@"\") + 1);
                        }
                        if (fileName.LastIndexOf(".") >= 0)
                        {
                            fileName = fileName.Substring(0, fileName.LastIndexOf("."));
                        }
                        ddata_.WriteImages(
                            this.fileNames_[i],
                            this.fps_,
                            this.imageWidth_,
                            this.imageHeight_,
                            this.textDraw_,
                            this.backgroundImageDraw_,
                            imgfmt,
                            txtSavePath.Text + @"\" + fileName);
                    }

                    this.AddStringToListBox(
                        lstMessage,
                        "> done.");
                }

                this.AddStringToListBox(
                    lstMessage,
                    "convert completed.");
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmConvert.btnConvert_Click]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
            finally
            {
                foreach (Control ictr in this.Controls)
                {
                    if (ictr is Button)
                    {
                        if (!ictr.Name.Equals("btnClose"))
                        {
                            ictr.Enabled = true;
                        }
                    }
                }
                processing_ = false;
                this.Cursor = Cursors.Default;
            }
        }

        private void btnSaveTo_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog dialog = new FolderBrowserDialog();
            dialog.ShowNewFolderButton = true;
            dialog.SelectedPath = txtSavePath.Text;
            if (dialog.ShowDialog() == DialogResult.OK)
            {
                txtSavePath.Text = dialog.SelectedPath;
            }
        }

        private void btnSelectFile_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog dlg = new OpenFileDialog();
                dlg.Filter = "*.wrn|*.wrn";
                dlg.Multiselect = true;
                if (dlg.ShowDialog() == DialogResult.OK)
                {
                    this.fileNames_.Clear();
                    lstSource.Items.Clear();

                    foreach (string iFileName in dlg.FileNames)
                    {
                        this.fileNames_.Add(iFileName);
                        lstSource.Items.Add(iFileName);
                    }
                }
            }
            catch (Exception exp)
            {
                MessageBox.Show(
                    "[frmConvert.btnSelectFile_Click]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
        }

        private void frmConvert_Load(object sender, EventArgs e)
        {
            this.Text = Application.ProductName + "[convert to image]";
        }

        #endregion  //  event
    }
}
